# -*- coding: utf-8 -*-
# @Time    : 2022/9/27 下午7:30
# @Author  : wzd
# @File    : __init__.py.py
"""
This module is used to calculate the potential energy
of molecules or complexes with metal atoms.

Used for molecules such as Metal organic cages.

"""