# -*- coding: utf-8 -*-
# @Time    : 2022/7/22 下午3:42
# @Author  : wzd
# @File    : __init__.py.py

"""
This module is used to calculate the potential energy
of pure organic molecules and complexes.

The potential energy is calculated by the force field
or empirical.

Based on:
-------------------------------------------------------
1. AMBER99
    Only contain the none-bonded interaction (vdw interaction and electrostatic force calculation).

2. Simple Lennard-Jones Potential

"""
